package be.jeffcheasey88.peeratcode.parser.java;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.lang.reflect.Modifier;
import java.util.LinkedList;
import java.util.List;

import be.jeffcheasey88.peeratcode.parser.Parser;
import be.jeffcheasey88.peeratcode.parser.Token;
import be.jeffcheasey88.peeratcode.parser.TokenType;
import be.jeffcheasey88.peeratcode.parser.Tokenizer;
import be.jeffcheasey88.peeratcode.parser.state.BuilderStateTree;
import be.jeffcheasey88.peeratcode.parser.state.InitialStateTree;
import be.jeffcheasey88.peeratcode.parser.state.RedirectStateTree;
import be.jeffcheasey88.peeratcode.parser.state.StateTree;

public class JavaParser  extends Parser<JavaElement> {
	
	public static long time;
	
	static {
		time = System.currentTimeMillis();
	}
	
	public static void main(String[] args) throws Exception{
		File file = new File("C:\\Users\\jeffc\\eclipse-workspace\\peer-at-code-backend\\src\\be\\jeffcheasey88\\peeratcode\\routes\\Result.java");
		BufferedReader reader = new BufferedReader(new FileReader(file));
		
		time = System.currentTimeMillis();
		
		Parser<JavaElement> parser = new JavaParser();
		JavaElement jFile = new JavaElement();
		parser.parse(reader, jFile);

		System.out.println((System.currentTimeMillis()-time)+"ms");
		
		parser.build(new BufferedWriter(new FileWriter(new File("/home/ParserV2.txt"))));
	}
	
	public JavaParser(){
		Tokenizer tokenizer = new Tokenizer();
		
		//SUMMARY
		//- Modifier
		//- Type
		//- Value
		//- Variable
		//- Function
		//- Class
		//- Import
		//- Package
		
		//MODIFIER
		StateTree<JavaElement> modifier = new StateTree<JavaElement>();
		modifier.then((validator) -> {
			boolean mod = false;
			while(validator.validate(
					(token) -> getModifier(token.getValue()) > 0, 
					(bag, token) -> {
						int m = getModifier(token.getValue());
						Integer current = bag.get();
						bag.set(current == null ? m : current+m);
					})) mod = true;
			return mod;
		}).end((a,b) -> a);
		
		//TYPE
		StateTree<JavaElement> type = new StateTree<JavaElement>();
		StateTree<JavaElement> type_ = type.then((validator) -> 
			validator.validate(
					(token) -> token.getType().equals(TokenType.NAME),
					(bag, token) -> bag.set(token)));
		StateTree<JavaElement> type_generic_begin = type_.then((validator) -> validator.validate(
				(token) -> token.getValue().equals("<"),
				(bag, token) -> bag.set(bag.<Token>get().concat(token))));
		StateTree<JavaElement> type_generic_name = type_generic_begin.then((validator) -> 
		validator.validate(
				(token) -> token.getType().equals(TokenType.NAME),
				(bag, token) -> bag.set(token)));
		StateTree<JavaElement> type_generic_split = type_generic_name.then((validator) -> validator.validate(
				(token) -> token.getValue().equals(","),
				(bag, token) -> bag.set(bag.<Token>get().concat(token))));
		StateTree<JavaElement> type_generic_end = type_generic_name.then((validator) -> validator.validate(
				(token) -> token.getValue().equals(">"),
				(bag, token) -> bag.set(bag.<Token>get().concat(token)))).loop();

		type_.end((a,b) -> a);
		type_generic_begin.then(type_generic_name);
		type_generic_split.then(type_generic_name);
		type_generic_end.then(type_generic_name);
		type_generic_end.then(type_generic_split);
		type_generic_end.end((a,b) -> a);
		
		//VALUE
		StateTree<JavaElement> value = new StateTree<>();
		StateTree<JavaElement> value_name = value.then((validator) -> validator.validate(
				(token) -> token.getType().equals(TokenType.NAME),
				(bag, token) -> {
					Token current = bag.get();
					if(current == null) current = token;
					else current = current.concat(token);
					bag.set(current);
				}));
		value_name.end((a,b) -> a);
		StateTree<JavaElement> value_call = value_name.then((validator) -> validator.validate((token) -> token.getValue().equals(".")));
		value_call.then(value_name);
		StateTree<JavaElement> value_arg_begin = value_name.then((validator) -> validator.validate((token) -> token.getValue().equals("(")));
		StateTree<JavaElement> value_arg_end = value_arg_begin.then((validator) -> validator.validate((token) -> token.getValue().equals(")")));
		value_arg_end.end((a,b) -> a);
		value_arg_end.then(value_call);
		StateTree<JavaElement> value_arg = value_arg_begin.then(new RedirectStateTree<>(value, (global, local) -> {
			Integer count = global.get("args");
			if(count == null) count = 0;
			global.set("arg"+count, local);
			global.set("args", (count+1));
		}));
		value_arg.then((validator) -> validator.validate((token) -> token.getValue().equals(",")))
			.then(value_arg);
		value.then((validator) -> {
			if(validator.validate((token) -> token.getValue().equals("'"))){
				validator.validate((token) -> token.getValue().equals("\\"));
				return
						validator.validate(
								(token) -> true,
								(bag, token) -> bag.set(token)) &&
						validator.validate((token) -> token.getValue().equals("'"));
			}
			return false;
		}).end((a,b) -> a);
		value.then((validator) -> {
			if(validator.validate((token) -> token.getValue().equals("\""))){
				
				while(validator.validate(
						(token) -> !token.getValue().equals("\""),
						(bag, token) -> {
							Token current = bag.get();
							if(current == null) current = token;
							else current = current.concat(token);
							bag.set(current);
						}));
				
				return validator.validate((token) -> token.getValue().equals("\""));
			}
			return false;
		}).end((a,b) -> a);
		
		//VARIABLE
		StateTree<JavaElement> variable = new StateTree<>();
		StateTree<JavaElement> variable_mod = variable.then(new RedirectStateTree<>(modifier, (global, local) -> global.set("modifier", local)));
		StateTree<JavaElement> variable_type = variable.then(new RedirectStateTree<>(type, (global, local) -> global.set("type", local)));
		variable_mod.then(variable_type);
		StateTree<JavaElement> variable_name = variable_type.then((validator) -> validator.validate(
				(token) -> token.getType().equals(TokenType.NAME),
				(bag, token) -> bag.set(token)));
		variable_name.then((validator) -> validator.validate((token) -> token.getValue().equals(";")))
			.end((a,b) -> new Variable(b));
		StateTree<JavaElement> variable_split = variable_name.then((validator) -> validator.validate((token) -> token.getValue().equals(",")));
		variable_split.then(variable_name);
		StateTree<JavaElement> variable_value = variable_name.then((validator) -> validator.validate((token) -> token.getValue().equals("=")))
			.then(new RedirectStateTree<>(value, (global, local) -> global.set("value", local)));
		variable_value.then(variable_split);
		variable_value.then((validator) -> validator.validate((token) -> token.getValue().equals(";")))
			.end((a,b) -> a);
		
		StateTree<JavaElement> function_container = new StateTree<>();
		
		//FUNCTION
		StateTree<JavaElement> function = new StateTree<>();
		StateTree<JavaElement> function_mod = function.then(new RedirectStateTree<>(modifier, (global, local) -> global.set("modifier", local)));
		StateTree<JavaElement> function_type = function.then(new RedirectStateTree<>(type, (global, local) -> global.set("type", local)));
		function_mod.then(function_type);
		function.then((validator) -> validator.validate((token) -> token.getValue().equals("static")) && validator.validate((token) -> token.getValue().equals("{"))).end((a,b) -> a).multiple(function_container);
		StateTree<JavaElement> function_name = function_type.then((validator) -> validator.validate(
				(token) -> token.getType().equals(TokenType.NAME),
				(bag, token) -> bag.set(token)));
		StateTree<JavaElement> function_begin = function_name.then((validator) -> validator.validate((token) -> token.getValue().equals("(")));
		StateTree<JavaElement> function_end = function_begin.then((validator) -> validator.validate((token) -> token.getValue().equals(")")));
		function_end.then((validator) -> validator.validate((token) -> token.getValue().equals("{"))).end((a,b) -> a).multiple(function_container);
		
		StateTree<JavaElement> function_throws = function_end.then((validator) -> validator.validate((token) -> token.getValue().equals("throws")))
				.then(new RedirectStateTree<>(type, (global, local) -> global.set(null)));
		function_throws.then((validator) -> validator.validate((token) -> token.getValue().equals("{"))).end((a,b) -> a).multiple(function_container);
		function_throws.then((validator) -> validator.validate((token) -> token.getValue().equals(","))).then(function_throws);
		
		
		StateTree<JavaElement> function_arg_type = function_begin.then(new RedirectStateTree<>(type, (global, local) -> global.set(null)));
		StateTree<JavaElement> function_arg_name = function_arg_type.then((validator) -> validator.validate((token) -> token.getType().equals(TokenType.NAME)));
		function_arg_name.then((validator) -> validator.validate((token) -> token.getValue().equals(","))).then(function_arg_type);
		function_arg_name.then(function_end);
		
		
		StateTree<JavaElement> clazz_container = new StateTree<>();
		clazz_container.then(variable);
		clazz_container.then(function);
		
		//CLASS
		StateTree<JavaElement> clazz_ = new StateTree<>();
		StateTree<JavaElement> clazz = new StateTree<>();
		
		StateTree<JavaElement> clazz_base = clazz.then((validator) -> validator.validate((token) -> token.getValue().equals("class")))
			.then(new RedirectStateTree<>(type, (global, local) -> global.set("name", local.get())));
		clazz_base.then((validator) -> validator.validate((token) -> token.getValue().equals("{")))
			.<JavaElement>end((JavaElement, bag) -> new Class(bag)).multiple(clazz_container);
		StateTree<JavaElement> clazz_implement = clazz_base.then((validator) -> validator.validate((token) -> token.getValue().equals("implements")));
		StateTree<JavaElement> clazz_implement_name = clazz_implement.then(new RedirectStateTree<>(type, (global, local) -> {
			Token token = global.get("implement");
			if(token == null) token = local.get();
			else token = token.concat(local.get());
			global.set("implement", token);
		}));
		clazz_implement_name.then((validator) -> validator.validate((token) -> token.getValue().equals(","))).then(clazz_implement_name);
		clazz_implement_name.then((validator) -> validator.validate((token) -> token.getValue().equals("{")))
			.<JavaElement>end((JavaElement, bag) -> new Class(bag)).multiple(clazz_container);
		StateTree<JavaElement> clazz_extend = clazz_base.then((validator) -> validator.validate((token) -> token.getValue().equals("extends")))
				.then(new RedirectStateTree<>(type, (global, local) -> global.set("extend", local)));
		clazz_extend.then((validator) -> validator.validate((token) -> token.getValue().equals("{")))
		.<JavaElement>end((JavaElement, bag) -> new Class(bag)).multiple(clazz_container);
		clazz_extend.then(clazz_implement);
		
		
		StateTree<JavaElement> clazz_mod = clazz_.then(new RedirectStateTree<>(modifier, (global, local) -> global.set("modifier", local)));
		clazz_mod.then(clazz);
		
		//IMPORT
		StateTree<JavaElement> importState = new StateTree<>();
		importState.then((validator) -> {
			if(validator.validate(
					(token) -> token.getValue().equals("import"),
					(bag, token) -> bag.set(new LinkedList<>()))){
				
				validator.validate(
						(token) -> token.getValue().equals("static"),
						(bag, token) -> bag.set("static", null));
				
				while(validator.validate(
						(token) -> !token.getValue().equals(";"),
						(bag, token) -> bag.<List<Token>>get().add(token))
					);
				
				return validator.validate((token) -> token.getValue().equals(";"));
			}
			return false;
		}).<JavaElement>end((javafile, bag) -> javafile);

		
		InitialStateTree<JavaElement> main = new InitialStateTree<>();
		
		//PACKAGE
		StateTree<JavaElement> pack = main.then((validator) -> {	
			if(validator.validate(
					(token) -> token.getValue().equals("package"),
					(bag, token) -> bag.set(new LinkedList<>()))){
				
				while(validator.validate(
						(token) -> !token.getValue().equals(";"),
						(bag, token) -> bag.<List<Token>>get().add(token))
					);
				return validator.validate((token) -> token.getValue().equals(";"));
			}
			return false;
		}).end((javafile, bag) -> javafile);
		main.multiple(pack);
		main.multiple(importState);
		main.multiple(clazz_);
		
		
		System.out.println((System.currentTimeMillis()-time)+"ms");
		
		setTokenizer(tokenizer);
		setStateTree(main);
	}
	
	public static int getModifier(String modifier){
		switch(modifier){
			case "public": return Modifier.PUBLIC;
			case "private": return Modifier.PRIVATE;
			case "protected": return Modifier.PROTECTED;
			case "static": return Modifier.STATIC;
			case "final": return Modifier.FINAL;
			case "synchronized": return Modifier.SYNCHRONIZED;
			case "volatile": return Modifier.VOLATILE;
			case "transient": return Modifier.TRANSIENT;
			case "native": return Modifier.NATIVE;
			case "abstract": return Modifier.ABSTRACT;
			case "strictfp": return Modifier.STRICT;
			default: break;
		}
		return 0;
	}
	
}
